@@ -24,12 +24,16 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <sys/socket.h>
+#include <netinet/in.h>
 #include <sys/un.h>
 #include "../util/mag_client.h"
 #include "../cspi/spi-private.h" /* A hack for now */
 
+#undef PRINT_TREE
+
 static void report_focus_event    (const AccessibleEvent *event, void *user_data);
 static void report_generic_event  (const AccessibleEvent *event, void *user_data);
+static void report_text_event     (const AccessibleEvent *event, void *user_data);
 static void report_button_press   (const AccessibleEvent *event, void *user_data);
 static void check_property_change (const AccessibleEvent *event, void *user_data);
 static SPIBoolean report_command_key_event  (const AccessibleKeystroke *stroke, void *user_data);
@@ -40,6 +44,10 @@ static int _festival_init ();
 static void _festival_say (const char *text, const char *voice, SPIBoolean shutup);
 static void _festival_write (const char *buff, int fd);
 
+#ifdef PRINT_TREE
+static void print_accessible_tree (Accessible *accessible, char *prefix);
+#endif
+
 static SPIBoolean use_magnifier = FALSE;
 static SPIBoolean use_festival = FALSE;
 static SPIBoolean festival_chatty = FALSE;
@@ -48,6 +56,7 @@ static AccessibleEventListener *focus_listener;
 static AccessibleEventListener *property_listener;
 static AccessibleEventListener *generic_listener;
 static AccessibleEventListener *button_listener;
+static AccessibleEventListener *text_listener;
 static AccessibleKeystrokeListener *command_key_listener;
 static AccessibleKeystrokeListener *ordinary_key_listener;
 static AccessibleKeySet            *command_keyset;
@@ -73,7 +82,7 @@ main (int argc, char **argv)
   modules = g_getenv ("GTK_MODULES");
   if (!modules || modules [0] == '\0')
     {
-      putenv ("GTK_MODULES=gail:atk-bridge");
+      putenv ("GTK_MODULES=");
     }
   modules = NULL;
 
@@ -82,6 +91,7 @@ main (int argc, char **argv)
   focus_listener = SPI_createAccessibleEventListener (report_focus_event, NULL);
   property_listener = SPI_createAccessibleEventListener (check_property_change, NULL); 
   generic_listener = SPI_createAccessibleEventListener (report_generic_event, NULL); 
+  text_listener = SPI_createAccessibleEventListener (report_text_event, NULL); 
   button_listener = SPI_createAccessibleEventListener (report_button_press, NULL);
   SPI_registerGlobalEventListener (focus_listener, "focus:");
   SPI_registerGlobalEventListener (property_listener, "object:property-change:accessible-selection"); 
@@ -89,7 +99,7 @@ main (int argc, char **argv)
   SPI_registerGlobalEventListener (generic_listener, "object:children-changed"); 
   SPI_registerGlobalEventListener (generic_listener, "object:visible-data-changed"); 
   SPI_registerGlobalEventListener (generic_listener, "object:text-selection-changed"); 
-  SPI_registerGlobalEventListener (generic_listener, "object:text-caret-moved"); 
+  SPI_registerGlobalEventListener (text_listener, "object:text-caret-moved"); 
   SPI_registerGlobalEventListener (generic_listener, "object:text-changed"); 
   SPI_registerGlobalEventListener (button_listener, "Gtk:GtkWidget:button-press-event");
   n_desktops = SPI_getDesktopCount ();
@@ -106,6 +116,9 @@ main (int argc, char **argv)
           application = Accessible_getChildAtIndex (desktop, j);
 	  s = Accessible_getName (application);
           fprintf (stderr, "app %d name: %s\n", j, s);
+#ifdef PRINT_TREE
+	  print_accessible_tree (application, "*");
+#endif
           SPI_freeString (s);
           Accessible_unref (application);
         }
@@ -169,11 +182,52 @@ get_environment_vars (void)
     }
 }
 
+#ifdef PRINT_TREE
+static void
+print_accessible_tree (Accessible *accessible, char *prefix)
+{
+	int n_children;
+	int i;
+	char *name;
+	char *role_name;
+	char *parent_name;
+	char *parent_role;
+	char child_prefix[100];
+	Accessible *child;
+	Accessible *parent;
+	
+	strncpy (child_prefix, prefix, 98);
+	strcat (child_prefix, "*");
+	parent = Accessible_getParent (accessible);
+	if (parent)
+	  {
+		parent_name = Accessible_getName (parent);
+		parent_role = Accessible_getRoleName (parent);
+	  }
+	Accessible_unref (parent);
+	name = Accessible_getName (accessible);
+	role_name = Accessible_getRoleName (accessible);
+	fprintf (stdout, "%sAccessible [%s] \"%s\"; parent [%s] %s\n",
+		 prefix, role_name, name, parent_role, parent_name);
+	SPI_freeString (name);
+	SPI_freeString (role_name);
+	SPI_freeString (parent_name);
+	n_children = Accessible_getChildCount (accessible);
+	for (i = 0; i < n_children; ++i)
+	        {
+			child = Accessible_getChildAtIndex (accessible, i);
+			print_accessible_tree (child, child_prefix);
+			Accessible_unref (child);
+	        }
+}
+#endif
+
 void
 report_focussed_accessible (Accessible *obj, SPIBoolean shutup_previous_speech)
 {
   char *s;
   int len;
+  long x, y, width, height;
 
   if (use_festival)
     {
@@ -192,14 +246,30 @@ report_focussed_accessible (Accessible *obj, SPIBoolean shutup_previous_speech)
   
   if (Accessible_isComponent (obj))
     {
-      long x, y, width, height;
       AccessibleComponent *component = Accessible_getComponent (obj);
       AccessibleComponent_getExtents (component, &x, &y, &width, &height,
                                       SPI_COORD_TYPE_SCREEN);
       fprintf (stderr, "Bounding box: (%ld, %ld) ; (%ld, %ld)\n",
                x, y, x+width, y+height);
+      if (Accessible_isText (obj))
+	{
+	  long x0, y0, xN, yN, w0, h0, wN, hN, nchars;
+	  AccessibleText *text = Accessible_getText (obj);
+	  nchars = AccessibleText_getCharacterCount (text);
+	  if (nchars > 0) 
+	    {
+	      AccessibleText_getCharacterExtents (text, 0, &x0, &y0, &w0, &h0,
+					      SPI_COORD_TYPE_SCREEN);
+	      AccessibleText_getCharacterExtents (text, nchars-1, &xN, &yN, &wN, &hN, 
+						  SPI_COORD_TYPE_SCREEN);
+	      x = MIN (x0, xN);
+	      width = MAX (x0 + w0, xN + wN) - x;
+	      fprintf (stderr, "Text bounding box: (%ld, %ld) ; (%ld, %ld)\n",
+		       x, y, x+width, y+height);
+	    }
+	}
       if (use_magnifier) {
-	      magnifier_set_roi ((short) 0, x, y, width, height);
+	magnifier_set_roi ((short) 0, x, y, width, height);
       }
     }
 
@@ -222,7 +292,7 @@ report_focussed_accessible (Accessible *obj, SPIBoolean shutup_previous_speech)
      text_interface = Accessible_getText (obj);
      first_sentence = AccessibleText_getTextAtOffset (
 	       text_interface, (long) 0, SPI_TEXT_BOUNDARY_SENTENCE_START, &start_offset, &end_offset);
-     if (first_sentence)
+     if (first_sentence && use_festival)
        {
 	 _festival_say(first_sentence, "voice_don_diphone", FALSE);
 	 SPI_freeString (first_sentence);
@@ -256,6 +326,23 @@ report_generic_event (const AccessibleEvent *event, void *user_data)
   fprintf (stderr, "%s event received\n", event->type);
 }
 
+void
+report_text_event (const AccessibleEvent *event, void *user_data)
+{
+  fprintf (stderr, "%s event received\n", event->type);
+  if (use_magnifier)
+    {
+      AccessibleText *text = Accessible_getText (event->source);
+      long offset = AccessibleText_getCaretOffset (text);
+      long x, y, w, h;
+      fprintf (stderr, "offset %d\n", (int) offset);
+      AccessibleText_getCharacterExtents (text, offset, &x, &y, &w, &h, 
+					  SPI_COORD_TYPE_SCREEN);
+      fprintf (stderr, "new roi %d %d %d %d\n", (int) x, (int) y, (int) w, (int) h);
+      magnifier_set_roi ((short) 0, x, y, w, h);
+    }
+}
+
 void
 report_button_press (const AccessibleEvent *event, void *user_data)
 {
@@ -311,6 +398,9 @@ simple_at_exit ()
   SPI_deregisterGlobalEventListenerAll (generic_listener);
   AccessibleEventListener_unref        (generic_listener);
 
+  SPI_deregisterGlobalEventListenerAll (text_listener);
+  AccessibleEventListener_unref        (text_listener);
+
   SPI_deregisterGlobalEventListenerAll (button_listener);
   AccessibleEventListener_unref        (button_listener);
 
