@@ -20,14 +20,18 @@
 #ifndef MAG_CLIENT_H_
 #define MAG_CLIENT_H_
 
+#include "util/Magnifier.h"
 
 #ifdef __cplusplus
 extern "C" {
 #endif /* __cplusplus */
 
-Accessibility_Magnifier get_magnifier(void);
+Accessibility_Magnifier get_magnifier (void);
 void magnifier_set_roi (int zoom_region, int x1, int y1, int x2, int y2);
 void magnifier_set_magnification (int zoom_region, float mag_factor_x, float mag_factor_y);
+void magnifier_resize_region (int zoom_region, int x1, int y1, int x2, int y2);
+int  magnifier_create_region (float zx, float zy, int x1, int y1, int x2, int y2);
+void magnifier_clear_all_regions (void);
 
 #ifdef __cplusplus
 }
